/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#ifndef CLIENT_H
#define CLIENT_H

#include "unlimited.h"

#ifdef CLIENT
/* control_ul, joy_ul and keybset_ul are old and should probably be deleted
soon */
enum control_ul { 
	COMPUTER, KEYBOARD, JOYSTICK
};
typedef enum control_ul CONTROL_UL;

struct joy_ul {
	Bool button[4];
	double posx, posy;	/* ranging from -10 to 10. y is accel/brake */
};
typedef struct joy_ul JOY_UL;

struct keybset_ul { /* This struct can be used both as hold the keys set by the
		   user as the keys pressed. Not sure if the type will be int,
	           it depends of how the keyboard will be read */
	uint32 left, right, accel, brake, gearup, geardown, clutch;
	uint32 changeview;
};
typedef struct keybset_ul KEYBSET_UL;
#endif

struct pref_ul {
	Bool		useobjf;	/* should object functions be used? */

#ifdef CLIENT	
	Bool 		kph; 		/* speed in kph? */
	Bool 		map;		/* show map? */
	Bool 		position;	/* show position? */
#endif
	Bool 		dltrack;	/* if don't have track, download it? */

	sInt		updatefreq;	/* updates/second */
	TIMESTAMP_UL	updatefreqts;
	uInt		maxupdates;	/* maximum # of updates; when reached,
					send a PRECISECONTROL_UL */
	
	uByte		servername[50];
	sInt 		serverport;

#ifdef CLIENT
	ENGPREF_UL 	engprefs;	/* engine preferences */
	sInt 		maxsoundcars;	/* max number of cars hearable */
	sInt 		maxfrontviewcars;/* max number of cars visible at front */
	sInt 		maxrearviewcars;/* max number of cars visible at rear */

	CONTROL_UL	controller;
	KEYBSET_UL	keybset;
	
	/* window stuff */
	sInt		width, height, depth;
	Bool		usedb, single;
	
	/* sound */
	Bool		sound;		/* use sound? */
	sInt		frequency;
	uByte		sounddevice[50];
	
	/* messages */
	sInt		maxmessages;
	uLong		messagetime;	/* in our timestamp format */
#endif
};
typedef struct pref_ul PREF_UL;

#ifdef CLIENT
/* font.c */
extern void setmessagetime	( uLong time );
extern void addmessage		( uByte *m, uShort from );
extern void cleanmessages	( void );
extern void initFont 		( void );
extern void showmessages	( void );
#endif

/* pref.c */
extern PREF_UL 	*prefs;
extern sInt loadprefs		( void );
extern void saveprefs		( void );

/* resp.c */
extern sInt doresp		( SERVERM_UL *resp );
extern void sendSpecialMessage	( CLIENTM_UL *request );

/* ulclient.c */
extern dFloat			frametime;
extern CLIENT_UL 		*own;
extern Bool			paused;
extern SIMPLECLIENT_UL 		*firstcli;
#ifdef CLIENT
extern FILE			*sound;
#endif

#endif
