/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "unlimited.h"

#define POLYN		10

static double 	gg[POLYN][POLYN], gf[POLYN], a[POLYN];

/* if orthogonal polynomes were used, it'd be a lot faster! */
static int g ( int i, double x, double y ) {
	switch ( i ) {
		case 0: 	  return 1;
		case 1: 	  return x;
		case 2: 	  return y;
		case 3: 	  return x*x;
		case 4: 	  return x*y;
		case 5: 	  return y*y;
		case 6: 	  return x*x*x;
		case 7: 	  return x*x*y;
		case 8: 	  return x*y*y;
		case 9: 	  return y*y*y;
	}
	return 0;	/* else gcc complains! */
}

static void makegg ( int n, int m, double x[n], double y[m] ) {
	int  i, j, k, l;
	double result = 0.0, nn = (double)n, mm = (double)m;

	for ( k = 0; k < POLYN; k++ )
		for ( l = 0; l < POLYN; l++ ) {
			for ( i = 0; i < nn; i++ ) /* inner product */
				for ( j = 0; j < mm; j++ )
					result += g( k, x[i]/nn, y[j]/mm ) * 
						g( l, x[i]/nn, y[j]/mm );
			gg[k][l] = result;
			result = 0.0;
		}
}

static void makegf ( int n, int m, double x[n], double y[m], double f[n][m] ) {
	int i, j, l;
	double result = 0.0;

	for ( l = 0; l < POLYN; l++ ) {
		for ( i = 0; i < n; i++ ) /* inner product */
			for ( j = 0; j < m; j++ )
				result += g( l, x[i]/(double)n, y[j]/(double)m )
						* f[i][j];
		gf[l] = result;
		result = 0.0;
	}
}

static double poly ( double x, double y ) {
	int i;
	double result = 0.0;
	for ( i = 0; i < POLYN; i++ )
		result += a[i]*g( i, x, y );
	return result;
}

/* given the arrays x[n], y[m], and f[n][m], this function finds the better
polynomial approximating the function, and interpolates stepx*stepy points in
[startx, endx]X[starty, endy], saving in output */
int lsq ( int n, int m, double x[n], double y[m], double f[n][m], double startx,
		double starty, double endx, double endy, int u, int v, 
		double output[u][v] ) {
	int i, j;
	double m1, m2, xn, yn;

	makegg( n, m, x, y ); /* inner products matrix */
	makegf( n, m, x, y, f );

	if ( linsys( POLYN, gg, gf, a ) == ERROR ) {
		fprintf( stderr, "Error, LSQ, undetermined linsys\n" );
		return ERROR;
	}

	m1 = (endx-startx)/u;
	m2 = (endy-starty)/v;
	
	for ( i = 0, xn = startx; i < u; i++, xn += m1 )
		for ( j = 0, yn = starty; j < v; j++, yn += m2 )
			output[i][j] = poly( xn, yn );
	return OK;
}
