/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

#ifndef UL_NET_H
#define UL_NET_H

#include "ulengine.h"
#include "time.h"

enum gamemode_ul {
	MATCH, FOREVER
};
typedef enum gamemode_ul GAMEMODE_UL;

/* message types, in both directions */
enum messagetype_ul {
	/* from client to server */
	UL_HELLO,

	/* from server to client */
	UL_WELCOME, UL_JOINED, UL_LEFT, UL_COUNTDOWN, UL_ENDRACE, UL_SMESSAGE,

	/* either way */
	UL_CMESSAGE, UL_PAUSE, UL_UNPAUSE, UL_QUIT, UL_GETFILE, UL_FILE,
	UL_CONTROL, UL_PRECISECONTROL, UL_COLLISION, UL_OBJECTCREATE,
	UL_OBJECTDESTROY, UL_OBJECTMODIFY, UL_OBJECTMOVE, UL_SPECIAL,
	UL_VARIABLE,

	/* and foo */
	UL_FOO
};
typedef enum messagetype_ul MESSAGETYPE_UL;

struct pause {
};
struct unpause {
};
struct quit {
};

struct cmessage {
	uShort		from, to;  	 /* client number */
	uByte		*message;
};

struct getfile {
	uByte		*name;
};			/* requests file */

struct file {
	uByte		*name;
	uLong		size;		/* so we can have \0s */
	uByte		*data;
};

struct control {
	uShort		number;
	sShort 		dx, dy, dz;	/* relative movement from
				   		last position */
	sShort 		dwx, dwy, dwz;	/* direction axis */
	sShort 		dangle;		/* rotation */
	uShort		timestamp;
	/* these above are all doubles, so the conversion is: 
		short/100 */
};

struct precisecontrol {
	uShort		number;
	VECTOR_UL	position;	/* last position */
	VECTOR_UL	direction;	/* direction */
	dFloat		angle;		/* rotation around dir */
	uShort		timestamp;
};

struct collision {
	uLong		id;
	VECTOR_UL	p;	/* where the collision happened? */
	uShort		timestamp;
};

struct objectcreate {
	uByte		*name;		/* this is were we get geometry.
					  Future versions should pass the
					  bounding sphere as object if
					  some client doesn't have it. */
	uLong		id;
	VECTOR_UL 	position;
	VECTOR_UL	direction;
	dFloat		angle;
	fFloat		mass;
	Bool		isstatic;
	uShort		timestamp;
};

struct objectdestroy {
	uLong		id;
	VECTOR_UL 	position;
	enum CAUSAMORTIS_UL { EXPLODE, VANISH, SHRINK, DISSOLVE } causamortis;
	uShort		timestamp;
};

struct objectmodify {
	uLong		id;
	VECTOR_UL 	position;
	VECTOR_UL	direction;
	dFloat		angle;
	uShort		timestamp;
};

struct objectmove {
	uLong		id;
	sShort 		dx, dy, dz;	/* relative movement from
				   		last position */
	sShort 		dwx, dwy, dwz;	/* direction axis */
	sShort 		dangle;		/* rotation */
	uShort		timestamp;
};

struct special {
	uShort	from, to;
	uByte	type;		/* message type */
	uByte 	*data;
};

struct variable {
	uByte	*name;
	uByte	*data;
};

/* requests -- messages from client to server */
union clientm_ul {
	MESSAGETYPE_UL request;

	struct {
		Bool		ai;
		uByte	*driver;  	 /* driver name */
		uByte	*model;
		ggi_color	color;		 /* color of user's client */
	}	hello;

	struct cmessage 	cmessage;
	struct pause		pause;
	struct unpause		unpause;
	struct quit		quit;
	struct getfile		getfile;
	struct file		file;
	struct control		control;
	struct precisecontrol	precisecontrol;
	struct collision	collision;
	struct objectcreate	objectcreate;
	struct objectdestroy	objectdestroy;
	struct objectmodify	objectmodify;
	struct objectmove	objectmove;
	struct special		special;
	struct variable		variable;
};
typedef union clientm_ul CLIENTM_UL;

/* responses -- messages from the server to the client */
union serverm_ul {
	MESSAGETYPE_UL response;

	struct {
		/* maybe we need a server mode field */
		uShort		number;		/* client's assigned # */
		uShort		racersn;	/* number of racers */
		uShort		objn;		/* number of objects */
	 	uByte 		*track;		/* name of track */
	 	uByte		*models;	/* \n-delimited list of vehicles */
	 	uByte		*drivers;	/* \n-delimited list of users */
		uByte		*objs;		/* object list */
		uByte		*mass;		/* separated by ; */
		uByte		*numbers;	/* separated by ; */
		uByte		*color;		/* format: r1,g1,b1:r2,g2... */
	}	welcome;

	struct {
	 	uByte 		*model;		/* vehicle */
	 	uByte 		*driver;	/* driver name */
		fFloat		mass;
		uShort		number;
		uByte		*color;
		Bool		ai;
	}	joinrace;

	struct {
		uLong		number;
	}	leftrace;
	
	struct {
	 	uShort		time;		/* time*10/0xFFFF */
	}	countdown;

	struct {
	}	end;

	struct {
		uShort		to;
		uByte 		*message;
	}	smessage;			/* message to everyone */

	struct cmessage 	cmessage;
	struct pause		pause;
	struct unpause		unpause;
	struct quit		quit;
	struct getfile		getfile;
	struct file		file;
	struct control		control;
	struct precisecontrol	precisecontrol;
	struct collision	collision;
	struct objectcreate	objectcreate;
	struct objectdestroy	objectdestroy;
	struct objectmodify	objectmodify;
	struct objectmove	objectmove;
	struct special		special;
	struct variable		variable;
};
typedef union serverm_ul SERVERM_UL;

#ifdef SERVER
struct update_ul {
	uLong			id;
	TIMESTAMP_UL		ts;
	VECTOR_UL		pos, dir;
	dFloat			angle;

	sInt			misses;		/* number of times we didn't
						  update, because it was too
						  far, or something. */
	struct update_ul	*next;
};
typedef struct update_ul UPDATE_UL;
#else
struct update_ul {
	TIMESTAMP_UL		time;
	VECTOR_UL		lastpos, lastdir;
	dFloat			lastangle;
	sInt			n;		/* number of updates to send
						a PRECISECONTROL_UL */
};
typedef struct update_ul UPDATE_UL;
#endif

/* netclient.c */
extern int			server;
extern void netflush		( void );
extern int netclient		( char *site_port );
extern void netclientsend	( CLIENTM_UL *msg );
extern SERVERM_UL *netclientrcv ( void );
extern int netclientmore 	( void );

/* netio.c */
extern void netclose		( int sock );
extern int netread 		( int sock, char *buf, int nbytes );
extern int netwrite 		( int sock, char *buf, int nbytes );

/* netlisten.c */
extern int netlisten 		( int port );
extern int netaccept 		( int lsock );

/* netopen.c */
extern int netopen		( char *site_port, int defport );

/* netselect.c */
extern void netaddsock 		( int sock );
extern void netdelsock 		( int sock );
extern int netwait 		( long msec );

#endif
