/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

/* This file, netclient.c, is adapted from original code by Steve Kirkendall.
Permission given. */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "unlimited.h"
#define SENDQ	16384
#define RECVQ	16384
#define SERVERN	"ulserver"

int server;

static unsigned char sendq[16384];
static int sendcnt;

/* flush outgoing messages, which were queued by netclientsend(). */
void netflush ( void ) {
	if (sendcnt > 0) {
		(void) netwrite( server, sendq, sendcnt );
		sendcnt = 0;
	}
}

/* create a server process */
static int forkserver( int port ) {
	fprintf( stderr, "Trying to create an server process\n" );

	switch ( fork() ) {
	  case -1: /* failed */
		perror( "fork()" );
		return ERROR;

	  case 0: /* child */
		/* exec the server.  Try a couple different paths */
		execlp( SERVERN, SERVERN, NULL );
		execl( ULDIR "server/ulserver", SERVERN, NULL);

		/* if we get here, then it failed */
		perror("exec(\"ulserver\")");
		return FATALERROR;

	  default: /* parent */
		/* cross your fingers.  The fork() worked, but will the
		 * execlp() ?
		 */
		 return OK;
	}
}	

/* Open a connection to a given server.  If the server doesn't exist and would
 * be local, then try to create one.
 */
int netclient ( char *site_port ) {
	int	i;

	/* try it the easy way first */
	server = netopen( site_port, DEFPORT );
	if ( server >= 0 )
		return server;

	/* would the server be on the local system? */
	if ( !strcmp( site_port, "localhost" ) ) {
		/* yes, using the default port */
		if ( forkserver( DEFPORT ) != OK )
			return FATALERROR;
	}
	else if ( !strncmp( site_port, "localhost:", 10 ) ) {
		/* yes, using an explicit port number */
		if ( forkserver( atoi( site_port + 10 ) ) != OK )
			return FATALERROR;
	}
	else
		/* no, damn! */
		return FATALERROR;

	/* try to connect to the server some more. */
	for (i = 0; i < 10 && server < 0; i++) {
		sleep(1);
		server = netopen( site_port, DEFPORT );
	}
	return server;
}

static void sendlong ( long l ) {
	sendq[sendcnt++] = l >> 24;
	sendq[sendcnt++] = l >> 16;
	sendq[sendcnt++] = l >> 8;
	sendq[sendcnt++] = l;
}

static void sendshort ( short l ) {
	sendq[sendcnt++] = l >> 8;
	sendq[sendcnt++] = l;
}

/* Send a message to the server opened by netclient() */
void netclientsend ( CLIENTM_UL *msg ) {
	int	i;

	/* if sendq is nearly full, then send it now */
	if ( sendcnt > SENDQ - 1024 )
		netflush();
/**** this previous lines probably will be deleted. */

	/* stuff the request type into the send queue */
	sendq[sendcnt++] = msg->request;

	/* convert the request to binary */
	switch ( msg->request ) {
	  case UL_HELLO:
		i = strlen(msg->hello.driver) + 1;
		memcpy( &sendq[sendcnt], msg->hello.driver, i );
		sendcnt += i;
		i = strlen(msg->hello.model) + 1;
		memcpy( &sendq[sendcnt], msg->hello.model, i );
		sendcnt += i;
		sendshort( msg->hello.color.r );
		sendshort( msg->hello.color.g );
		sendshort( msg->hello.color.b );
		break;
		
	  case UL_QUIT:
	  case UL_PAUSE:
	  case UL_UNPAUSE:
		break;

	  case UL_CMESSAGE:
		sendshort( msg->cmessage.from );
		sendshort( msg->cmessage.to );
		i = strlen(msg->cmessage.message) + 1;
		memcpy( &sendq[sendcnt], msg->cmessage.message, i );
		sendcnt += i;
		break;

	  case UL_GETFILE:
		i = strlen(msg->getfile.name) + 1;
		memcpy( &sendq[sendcnt], msg->getfile.name, i );
		sendcnt += i;
		break;

	  case UL_FILE:
		i = strlen(msg->file.name) + 1;
		memcpy( &sendq[sendcnt], msg->file.name, i );
		sendcnt += i;
		sendlong( msg->file.size );
		i = strlen(msg->file.data) + 1;
		if ( i + sendcnt < SENDQ ) {
			memcpy(&sendq[sendcnt], msg->file.data, i);
			sendcnt += i;
		}
		else {
			netflush();
			netwrite( server, msg->file.data, i );
		}
		break;

	  case UL_CONTROL:
		sendshort( msg->control.dx );	sendshort( msg->control.dy );
		sendshort( msg->control.dz );	sendshort( msg->control.dwx );
		sendshort( msg->control.dwy );	sendshort( msg->control.dwz );
		sendshort( msg->control.dangle );
		sendshort( msg->control.timestamp );
		break;

	  case UL_PRECISECONTROL:
		sendlong( (long)(msg->precisecontrol.direction.x*1000) );
		sendlong( (long)(msg->precisecontrol.direction.y*1000) );
		sendlong( (long)(msg->precisecontrol.direction.z*1000) );
		sendlong( (long)(msg->precisecontrol.position.x*1000) );
		sendlong( (long)(msg->precisecontrol.position.y*1000) );
		sendlong( (long)(msg->precisecontrol.position.z*1000) );
		sendlong( (long)(msg->precisecontrol.angle*1000) );
		sendshort( msg->precisecontrol.timestamp );
		break;

	  case UL_COLLISION:
		sendlong( msg->collision.id );
		/**** tests to see if the values are out of bounds should be made */
		sendlong( (long)msg->collision.p.x*1000 );
		sendlong( (long)msg->collision.p.y*1000 );
		sendlong( (long)msg->collision.p.z*1000 );
		sendshort( msg->collision.timestamp );
		break;

	  case UL_OBJECTCREATE:
		if ( general.createobject ) {
			i = strlen(msg->objectcreate.name) + 1;
			memcpy( &sendq[sendcnt], msg->objectcreate.name, i );
			sendcnt += i;
			sendlong( msg->objectcreate.id );
			sendlong( (long)(msg->objectcreate.position.x*1000) );
			sendlong( (long)(msg->objectcreate.position.y*1000) );
			sendlong( (long)(msg->objectcreate.position.z*1000) );
			sendlong( (long)(msg->objectcreate.direction.x*1000) );
			sendlong( (long)(msg->objectcreate.direction.y*1000) );
			sendlong( (long)(msg->objectcreate.direction.z*1000) );
			sendlong( (long)(msg->objectcreate.angle*1000) );
			sendlong( (long)(msg->objectcreate.mass*1000) );
			sendshort( msg->objectcreate.timestamp );
		}
		else
#ifdef CLIENT
			addmessage( "Object creation not allowed", 0 );
#else
			fprintf( stderr, "Object creation not allowed" );
#endif
		break;
		
	  case UL_OBJECTDESTROY:
		if ( general.destroyobject ) {
			sendlong( msg->objectdestroy.id );
			sendlong( (long)(msg->objectdestroy.position.x*1000) );
			sendlong( (long)(msg->objectdestroy.position.y*1000) );
			sendlong( (long)(msg->objectdestroy.position.z*1000) );
			sendq[sendcnt++] = msg->objectdestroy.causamortis;
			sendshort( msg->objectdestroy.timestamp );
		}
		else
#ifdef CLIENT
			addmessage( "Object destruction not allowed", 0 );
#else
			fprintf( stderr, "Object destruction not allowed" );
#endif
		break;

	  case UL_OBJECTMODIFY:
		if ( general.modifyobject ) {
			sendlong( msg->objectmodify.id );
			sendlong( (long)(msg->objectmodify.position.x*1000) );
			sendlong( (long)(msg->objectmodify.position.y*1000) );
			sendlong( (long)(msg->objectmodify.position.z*1000) );
			sendlong( (long)(msg->objectmodify.direction.x*1000) );
			sendlong( (long)(msg->objectmodify.direction.y*1000) );
			sendlong( (long)(msg->objectmodify.direction.z*1000) );
			sendlong( (long)(msg->objectmodify.angle*1000) );
			sendshort( msg->objectmodify.timestamp );
		}
		else
#ifdef CLIENT
			addmessage( "Object modification not allowed", 0 );
#else
			fprintf( stderr, "Object modification not allowed" );
#endif
		break;

	  case UL_OBJECTMOVE:
		if ( general.modifyobject ) {
			sendlong( msg->objectmove.id );
			sendshort( msg->objectmove.dx );
			sendshort( msg->objectmove.dy );
			sendshort( msg->objectmove.dz );
			sendshort( msg->objectmove.dwx );
			sendshort( msg->objectmove.dwy );
			sendshort( msg->objectmove.dwz );
			sendshort( msg->objectmove.dangle );
			sendshort( msg->objectmove.timestamp );
		}
		else
#ifdef CLIENT
			addmessage( "Object modification not allowed", 0 );
#else
			fprintf( stderr, "Object modification not allowed" );
#endif
		break;

	  case UL_SPECIAL:
		sendshort( msg->special.from );
		sendshort( msg->special.to );
		sendq[sendcnt++] = msg->special.type;
		i = strlen(msg->special.data) + 1;
		memcpy( &sendq[sendcnt], msg->special.data, i );
		sendcnt += i;
		break;

	  default:
		fprintf(stderr, "Invalid request %d\n", msg->request);
	}
}

static char 	recvq[RECVQ];
static int	recvcnt;
static int	recvused;

/* Test whether all of the required fields are present.  Return 1 if missing,
 * or 0 if complete.
 */
static int incomplete ( int bytes, int strings ) {
	/* is it obviously too short? */
	if ( recvused + bytes + strings > recvcnt )
		return 1;

	/* if no strings needed, then we got it */
	if ( strings == 0 )
		return 0;

	/* count the strings */
	for ( ; recvused + bytes < recvcnt; bytes++ )
		if ( recvq[recvused + bytes] == '\0' && --strings == 0 )
			return 0;

	/* ran out of bytes before the last string */
	return 1;
}


static long recvlong( void ) {
	long	l;

	l = ((((long)recvq[recvused] << 24) & 0xff000000)
	   | (((long)recvq[recvused + 1] << 16) & 0xff0000)
	   | (((long)recvq[recvused + 2] << 8) & 0xff00)
	   | (((long)recvq[recvused + 3] & 0xff)));
	recvused += 4;
	return l;
}

static short recvshort( void ) {
	short	l;

	l = ((((short)recvq[recvused + 2] << 8) & 0xff00)
	   | (((short)recvq[recvused + 3] & 0xff)));
	recvused += 2;
	return l;
}

/* Receive a message from the server opened by netclient().  It is assumed
 * that either netselect indicated there was a message available, or you want
 * to wait for one now.  Any (char *) fields in the response will point to
 * statically allocated space --  you don't need to free it, but you should
 * make your own copy of it before the next call to netclientrcv().
 */
SERVERM_UL *netclientrcv ( void ) {
	static SERVERM_UL response;
	int		beginning;
	int		i;

	/* if no bytes in recvq, then read some */
	if ( recvused == recvcnt ) {
		recvcnt = netread ( server, recvq, RECVQ);
		if (recvcnt <= 0)
			return NULL;
		recvused = 0;
	}

	/* We may have an incomplete message.  Do some preparation so we can
	 * read some more and start over if necessary.
	 */
	beginning = recvused;

	for (;;) {
		/* parse the response */
		response.response = recvq[recvused++];
		switch ( response.response ) {
		  case UL_WELCOME:
			if ( incomplete( 2, 6 ) )
				break;
			response.welcome.number = recvshort();
			response.welcome.track = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			response.welcome.models = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			response.welcome.drivers = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			response.welcome.mass = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			response.welcome.numbers = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			response.welcome.color = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			return &response;

		  case UL_JOINED:
			if ( incomplete( 7, 3 ) )
				break;
			response.joinrace.model = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			response.joinrace.driver = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			response.joinrace.mass = ((double)recvlong())/1000;
			response.joinrace.number = recvshort();
			response.joinrace.color = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			response.joinrace.ai = (Bool)recvq[recvused++];
			return &response;

		  case UL_LEFT:
			if ( incomplete( 2, 0 ) )
				break;
			response.leftrace.number = recvshort();
			return &response;

		  case UL_COUNTDOWN:
			if ( incomplete( 2, 0 ) )
				break;
			response.countdown.time = recvshort();
			return &response;

		  case UL_ENDRACE:
		  case UL_QUIT:
		  case UL_PAUSE:
		  case UL_UNPAUSE:
			return &response;

		  case UL_SMESSAGE:
			if ( incomplete( 0, 1 ) ) 
				break;
			response.smessage.to = recvshort();
			response.smessage.message = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			return &response;

		  case UL_CMESSAGE:
			if ( incomplete( 2, 1 ) ) 
				break;
			response.cmessage.from = recvshort();
			response.cmessage.to = recvshort();
			response.cmessage.message = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			return &response;

		  case UL_CONTROL:
			if ( incomplete( 16, 0 ) )
				break;
			response.control.number = recvshort();
			response.control.dx = recvshort();
			response.control.dy = recvshort();
			response.control.dz = recvshort();
			response.control.dwx = recvshort();
			response.control.dwy = recvshort();
			response.control.dwz = recvshort();
			response.control.dangle = recvshort();
			response.control.timestamp = recvshort();
			return &response;

		  case UL_PRECISECONTROL:
			if ( incomplete( 16, 0 ) )
				break;
			response.precisecontrol.number = recvshort();
			response.precisecontrol.position.x = ((double)recvlong()/1000);
			response.precisecontrol.position.y = ((double)recvlong()/1000);
			response.precisecontrol.position.z = ((double)recvlong()/1000);
			response.precisecontrol.direction.x = ((double)recvlong()/1000);
			response.precisecontrol.direction.y = ((double)recvlong()/1000);
			response.precisecontrol.direction.z = ((double)recvlong()/1000);
			response.precisecontrol.angle = ((double)recvlong()/1000);
			response.precisecontrol.timestamp = recvshort();
			return &response;

		  case UL_GETFILE:
			if ( incomplete( 0, 1 ) )
				break;
			response.getfile.name = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			return &response;

		  case UL_FILE: /***** not finished */
#ifdef NOTDONE
			if ( incomplete( 4, 2 )) {
				/* It might not fit in recvq[].  We'd better
				 * receive the string piecemeal, and assemble
				 * it using netbuild().
				 */
/******				netbuild( NULL ); gotta fix this */
				do {
					response.track.data =
						nrbuildn(&recvq[recvused],
							recvcnt - recvused);
					recvcnt = nrread(nrserver, recvq, RECVQ);
					recvused = 0;
					if (recvcnt <= 0)
						return NULL;
				} while (incomplete(0, 1));
				i = strlen(&recvq[recvused]) + 1;
				response.file.data = nrbuildn(&recvq[recvused], i);
				recvused += i;
			}
			else {
				/* We got lucky.  It fits */
				response.file.data = &recvq[recvused];
				while (recvq[recvused++]) { }
			}
			return &response;
#endif
		  case UL_COLLISION:
			if ( incomplete( 18, 0 ) ) 
				break;
			response.collision.id = recvlong();
			response.collision.p.x = ((double)recvlong())*0.001;
			response.collision.p.y = ((double)recvlong())*0.001;
			response.collision.p.z = ((double)recvlong())*0.001;
			response.collision.timestamp = recvshort();
			break;

		  case UL_OBJECTCREATE:
			if ( incomplete( 38, 1 ) ) 
				break;
			if ( general.createobject == False ) {
				/* this should never occur! */
				/* print a message on client's screen */
				return NULL;
			}
			response.objectcreate.name = &recvq[recvused];
			while (recvq[recvused++]) { }
			response.objectcreate.id = recvlong();
			response.objectcreate.position.x = ((double)recvlong())*0.001;
			response.objectcreate.position.y = ((double)recvlong())*0.001;
			response.objectcreate.position.z = ((double)recvlong())*0.001;
			response.objectcreate.direction.x = ((double)recvlong())*0.001;
			response.objectcreate.direction.y = ((double)recvlong())*0.001;
			response.objectcreate.direction.z = ((double)recvlong())*0.001;
			response.objectcreate.angle = ((double)recvlong())*0.001;
			response.objectcreate.mass = ((float)recvlong())*0.001;
			response.objectcreate.timestamp = recvshort();
			return &response;

		  case UL_OBJECTDESTROY:
			if ( incomplete( 19, 0 ) ) 
				break;
			if ( general.destroyobject == False ) {
				/* this should never occur! */
				/* print a message on client's screen */
				return NULL;
			}
			response.objectdestroy.id = recvlong();
			response.objectdestroy.position.x = ((double)recvlong())*0.001;
			response.objectdestroy.position.y = ((double)recvlong())*0.001;
			response.objectdestroy.position.z = ((double)recvlong())*0.001;
			response.objectdestroy.causamortis = recvq[recvused];
			response.objectdestroy.timestamp = recvshort();
			return &response;

		  case UL_OBJECTMODIFY:
			if ( incomplete( 30, 0 ) ) 
				break;
			if ( general.modifyobject == False ) {
				/* this should never occur! */
				/* print a message on client's screen */
				return NULL;
			}
			response.objectmodify.position.x = ((double)recvlong())*0.001;
			response.objectmodify.position.y = ((double)recvlong())*0.001;
			response.objectmodify.position.z = ((double)recvlong())*0.001;
			response.objectmodify.direction.x = ((double)recvlong())*0.001;
			response.objectmodify.direction.y = ((double)recvlong())*0.001;
			response.objectmodify.direction.z = ((double)recvlong())*0.001;
			response.objectmodify.angle = ((double)recvlong())*0.001;
			response.objectmodify.timestamp = recvshort();
			return &response;

		  case UL_OBJECTMOVE:
			if ( incomplete( 22, 0 ) ) 
				break;
			if ( general.modifyobject == False ) {
				/* this should never occur! */
				/* print a message on client's screen */
				return NULL;
			}
			response.objectmove.id = recvlong();
			response.objectmove.dx = recvshort();
			response.objectmove.dy = recvshort();
			response.objectmove.dz = recvshort();
			response.objectmove.dwx = recvshort();
			response.objectmove.dwy = recvshort();
			response.objectmove.dwz = recvshort();
			response.objectmove.dangle = recvshort();
			response.objectmove.timestamp = recvshort();
			return &response;

		  case UL_SPECIAL:
			if ( incomplete( 5, 1 ) )
				break;
			response.special.from = recvshort();
			response.special.to = recvshort();
			response.special.type = recvq[recvused++];
			response.special.data = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			return &response;

		  default:
			fprintf(stderr, "unknown response %d\n",
							response.response);
			return NULL;
		}

		/* If we get here, then we had an incomplete message at the
		 * end of recvq[].  Try to read the rest of it.
		 */
		if (beginning != 0) {
			memcpy( recvq, &recvq[beginning], recvcnt - beginning );
			recvcnt -= beginning;
		}
		beginning = recvused = 0;
		i = netread( server, &recvq[recvcnt], RECVQ - recvcnt );
		if (i <= 0)
			return NULL;
		recvcnt += beginning;
	}
/*NOTREACHED*/
}


/* Return TRUE if there is at least a partial message in the receive queue */
int netclientmore ( void ) {
	return (recvused < recvcnt);
}
